/*
 *  Patches a binary file with another binary file
 */

#include <stdio.h>
#include <stdlib.h>

long getFilesize(FILE *input_file);

int main (int argc, char *argv[])
{
    switch (argc) {
    case 4:
        return(binpatcher_func(argv[1], atoi(argv[2]), argv[3]));
        break;
    default:
        printf("Usage:\n");
        printf("arg1 = file name (recipient)\n");
        printf("arg2 = patch offset\n");
        printf("arg3 = file name (patch)\n");
        return(EXIT_FAILURE);
        break;
    }
}

int binpatcher_func(char *recipient_name, long offset,  char *patch_name) {
    FILE *recipient_file, *patch_file;
    long recipient_size, patch_size;
    unsigned char *patch;

    /* check if the recipient and patch exist */
    recipient_file = fopen(recipient_name,"rb");
    if (!recipient_file) {
        printf("The file %s doesn't exist\n", recipient_name);
        return(EXIT_FAILURE);
    }
    patch_file = fopen(patch_name,"rb");
    if (!patch_file) {
        printf("The file %s doesn't exist\n", patch_name);
        return(EXIT_FAILURE);
    }
    fclose(recipient_file);
    fclose(patch_file);

    /* open the recipient and patch files for reading and writing */
    if (!(recipient_file=fopen(recipient_name,"r+b"))) {
        printf("Couldn't open %s\n", recipient_name);
        return(EXIT_FAILURE);
    }
    patch_file = fopen(patch_name,"r+b");
    if (!(patch_file=fopen(patch_name,"r+b"))) {
        printf("Couldn't open %s\n", patch_name);
        return(EXIT_FAILURE);
    }

    /* check if the recipient is patchable */
    recipient_size=getFilesize(recipient_file);
    patch_size=getFilesize(patch_file);

    if (offset < 0) {
        printf("Patch start < 0\n");
        fclose(recipient_file);
        fclose(patch_file);
        return(EXIT_FAILURE);
    }
    else if (offset + patch_size > recipient_size) {
        printf("Patch end > file size\n");
        fclose(recipient_file);
        fclose(patch_file);
        return(EXIT_FAILURE);
    }

    /* read the patch into memory */
    patch = (unsigned char *) malloc(sizeof(unsigned char) * patch_size);
    if (!patch) {
        printf("Couldn't allocate memory\n");
        fclose(recipient_file);
        fclose(patch_file);
        return(EXIT_FAILURE);
    }
    fread(patch, sizeof(unsigned char), patch_size, patch_file);
    fclose(patch_file);

    /* patch the recipient */
    fseek(recipient_file, offset, SEEK_SET);
    fwrite(patch, sizeof(unsigned char), patch_size, recipient_file);
    fclose(recipient_file);
    free(patch);
    return(EXIT_SUCCESS);
}

/*
 * Returns the size of a file, without moving the file pointer.
 * The file must already be open when this function is called.
 */
long getFilesize(FILE *input_file)
{
    long file_length, file_pos;

    file_pos = ftell(input_file);     /* get the current file position */
    fseek(input_file, 0, SEEK_END);   /* move file pointer to file end */
    file_length = ftell(input_file);
    fseek(input_file, 0, file_pos);   /* restore the original file position */
    return(file_length);
}
