; =================================================
; German keyboard utility for Ultima 3 and Ultima 4
; Assembled with NASM 0.98.38 (as a COM file)
; Last update: 23-March-2004
;
; How to assemble:
; nasmw -fbin -o keyb_de.com keyb_de.asm
; =================================================

                org     100h

section         .text
                ; ------------------------------------------
                ;     move stack and resize memory block
                ; ------------------------------------------
                mov     sp,stack_end

                mov     bx,stack_end+0fh
                mov     cl,4
                shr     bx,cl
                mov     ah,4ah
                int     21h

                ; --------------------------
                ;     parse command line
                ; --------------------------
ws_node:
                mov     bx,[chars_read]
                ; chars left?
                cmp     bl, byte [80h]
                jae     end_node
                ; is the next char a whitespace?
                cmp     byte [80h+1+bx],20h ; space
                jz      is_ws
                cmp     byte [80h+1+bx],9 ; tab
                jz      is_ws
                ; begin new argument
                mov     ax,[chars_read]
                mov     word [start_ofs],ax
                mov     word [length],1
                inc     word [chars_read]
                inc     word [non_empty]
                jmp     non_ws_node
    is_ws:
                inc     word [chars_read]
                jmp     ws_node

non_ws_node:
                mov     bx,[chars_read]
                ; chars left?
                cmp     bl, byte [80h]
                jae     end_node
                ; is the next char a non-whitespace?
                cmp     byte [80h+1+bx],20h ; space
                jz      end_node
                cmp     byte [80h+1+bx],9 ; tab
                jz      end_node
                ; it's a non-whitespace
                inc     word [length]
                inc     word [chars_read]
                jmp     non_ws_node
end_node:
                ; ----------------------------------------
                ;     prepare command line for program
                ; ----------------------------------------
                cmp     word [non_empty],0
                jnz     cl_nonempty

                mov     dx,usage_str
                mov     ah,9
                int     21h
                jmp     quit_simple
cl_nonempty:
                ; copy file name
                mov     cx,[length]
                mov     si,[start_ofs]
                add     si,81h
                mov     di,file_name
                cld
                rep     movsb
                mov     byte [di],0   ; trailing zero

                ; copy rest of command line
                mov     si,[start_ofs]
                add     si,[length]   ; si = offset of the first char after the first param (whitespace or 0dh)

                mov     cl,[80h]
                mov     ch,0
                sub     cx,si   ; cx = number of characters after the first param
                add     si,81h
                mov     di,cmd_line+1
                cld
                mov     byte [cmd_line],cl
                rep     movsb
                mov     byte [di],0dh

                ; -----------------------------------------
                ;     int 16h vector --> keyboard patch
                ; -----------------------------------------
                mov     ax,3516h   ; get old int 16h
                int     21h
                mov     [int_16h_ofs],bx
                mov     [int_16h_seg],es

                mov     ax,2516h   ; set new int 16h
                mov     dx,cs
                mov     ds,dx
                mov     dx,keyboard_patch
                int     21h

                ; -------------------
                ;     run program
                ; -------------------
run_program:
                mov     word [seg_env],0   ; copy environment from parent process
                mov     word [cmd_line_ofs],cmd_line
                mov     word [cmd_line_seg],cs
                mov     word [fcb_1_ofs],5Ch
                mov     word [fcb_1_seg],cs
                mov     word [fcb_2_ofs],6Ch
                mov     word [fcb_2_seg],cs

                mov     bx,cs
                mov     es,bx
                mov     bx,seg_env   ; es:bx = pointer to control block

                mov     dx,cs
                mov     ds,dx
                mov     dx,file_name   ; ds:dx = pointer to path name

                mov     ax,4b00h
                int     21h

                ; restore segment registers and stack
                cli
                mov     sp,cs
                mov     ss,sp
                mov     ds,sp
                mov     es,sp
                mov     sp,stack_end
                sti

                jnc     restore_16h

                ; an error occurred
                mov     dx,error_str
                mov     ah,9
                int     21h

                ; ----------------------------------------------
                ;     int 16h vector --> old int 16h handler
                ; ----------------------------------------------
restore_16h:
                mov     dx,[int_16h_ofs]
                mov     ds,[int_16h_seg]
                mov     ax,2516h
                int     21h

quit_simple:
                mov     ah,4ch
                int     21h

; -----------------------------------------------------------------------------------------------
;     patch: checks for un-American scancode/ascii combinations returned by a German keyboard
; -----------------------------------------------------------------------------------------------
keyboard_patch:
    ; old int 16h handler
    mov     [cs:func_code],ah
    pushf
    call    far [cs:int_16h_ofs]

    ; save the flags
    pushf
    pop     word [cs:flags]

    ; modify ax, if necessary
    pushf

    ; check if we want to modify the output
    cmp     byte [cs:func_code],0
    jz      modify_output
    cmp     byte [cs:func_code],10h
    jz      modify_output
    test    word [cs:flags],40h     ; zero flag
    jnz     keymod_end
    cmp     byte [cs:func_code],1
    jz      modify_output
    cmp     byte [cs:func_code],11h
    jz      modify_output
    jmp     keymod_end

modify_output:
    cmp     ax,2c79h   ; y pressed on german kb with german kb driver
    jz      y_pressed
    cmp     ax,157ah   ; z pressed on german kb with german kb driver
    jz      z_pressed
    jmp     keymod_end
y_pressed:
    mov     ax,1579h   ; ah = us_scancode(y)
    jmp     keymod_end
z_pressed:
    mov     ax,2c7ah   ; ah = us_scancode(z)
keymod_end:
    popf
    retf    2

; -----------------------------
;     initialized variables
; -----------------------------
non_empty       dw      0
chars_read      dw      0
start_ofs       dw      0
length          dw      0
usage_str       db      'Usage:',0ah,0dh,'keyb_de filename.ext [parameters]',0ah,0dh,'$'
error_str       db      'An error occurred.',0ah,0dh,'$'
func_code       db      0
flags           dw      0

; -------------------------------
;     uninitialized variables
; -------------------------------
section         .bss

file_name       resb    128     ; program name + trailing zero
cmd_line        resb    128     ; command line passed to program

seg_env         resw    1
cmd_line_ofs    resw    1
cmd_line_seg    resw    1
fcb_1_ofs       resw    1
fcb_1_seg       resw    1
fcb_2_ofs       resw    1
fcb_2_seg       resw    1

int_16h_ofs     resw    1
int_16h_seg     resw    1

stack_start     resb    256
stack_end:
